@@ -69,6 +69,9 @@ static const char *cluster_config_options[] = {
 	"connection_lifetime",
 	"query_timeout",
 	"disable_binary",
+	"keepalive_idle",
+	"keepalive_interval",
+	"keepalive_count",
 	NULL
 };
 
@@ -257,6 +260,12 @@ set_config_key(ProxyFunction *func, ProxyConfig *cf, const char *key, const char
 		cf->query_timeout = atoi(val);
 	else if (pg_strcasecmp("disable_binary", key) == 0)
 		cf->disable_binary = atoi(val);
+	else if (pg_strcasecmp("keepalive_idle", key) == 0)
+		cf->keepidle = atoi(val);
+	else if (pg_strcasecmp("keepalive_interval", key) == 0)
+		cf->keepintvl = atoi(val);
+	else if (pg_strcasecmp("keepalive_count", key) == 0)
+		cf->keepcnt = atoi(val);
 	else
 		plproxy_error(func, "Unknown config param: %s", key);
 }
@@ -391,6 +400,28 @@ extract_part_num(const char *partname, int *part_num)
 	return false;
 }
 
+/*
+ * Validate single cluster option
+ */
+static void
+validate_cluster_option(const char *name, const char *arg)
+{
+	const char **opt;
+
+	/* see that a valid config option is specified */
+	for (opt = cluster_config_options; *opt; opt++)
+	{
+		if (pg_strcasecmp(*opt, name) == 0)
+			break;
+	}
+
+	if (*opt == NULL)
+		elog(ERROR, "Pl/Proxy: invalid server option: %s", name);
+	else if (strspn(arg, "0123456789") != strlen(arg))
+		elog(ERROR, "Pl/Proxy: only integer options are allowed: %s=%s",
+			 name, arg);
+}
+
 /*
  * Validate the generic option given to servers or user mappings defined with
  * plproxy foreign data wrapper.  Raise an ERROR if the option or its value is
@@ -435,20 +466,7 @@ plproxy_fdw_validator(PG_FUNCTION_ARGS)
 			}
 			else
 			{
-				const char **opt;
-
-				/* see that a valid config option is specified */
-				for (opt = cluster_config_options; *opt; opt++)
-				{
-					if (pg_strcasecmp(*opt, def->defname) == 0)
-						break;
-				}
-
-				if (*opt == NULL)
-					elog(ERROR, "Pl/Proxy: invalid server option: %s", def->defname);
-				else if (strspn(arg, "0123456789") != strlen(arg))
-					elog(ERROR, "Pl/Proxy: only integer options are allowed: %s=%s",
-						 def->defname, arg);
+				validate_cluster_option(def->defname, arg);
 			}
 		}
 		else if (catalog == UserMappingRelationId)
@@ -465,8 +483,7 @@ plproxy_fdw_validator(PG_FUNCTION_ARGS)
 		}
 		else if (catalog == ForeignDataWrapperRelationId)
 		{
-			 /* At the moment there are no options to the fdw itself */
-			 elog(WARNING, "Pl/Proxy: foreign data wrapper takes no options");
+			validate_cluster_option(def->defname, arg);
 		}
 	}
 
@@ -553,6 +570,16 @@ reload_sqlmed_cluster(ProxyFunction *func, ProxyCluster *cluster,
 		appendStringInfo(user_options, "%s='%s' ", def->defname, strVal(def->arg));
 	}
 
+	/*
+	 * Collect the configuration definitions from foreign data wrapper.
+	 */
+	foreach(cell, fdw->options)
+	{
+		DefElem    *def = lfirst(cell);
+
+		set_config_key(func, &cluster->config, def->defname, strVal(def->arg));
+	}
+
 	/*
 	 * Collect the cluster configuration and partition definitions from foreign
 	 * server options. At first pass just collect the cluster options and count
